import requests
import gradio as gr
import json
import time

# 定義 Ollama 的 API 端點與模型名稱
OLLAMA_API_URL = "http://localhost:11434/api/generate"  # 預設端點
MODEL_NAME = "qwen2.5:3b"  # 替換為您已安裝的模型名稱
this_turn=""
custom_css = """
#scrollable-markdown {
    max-height: 500px; /* 設定最大高度 */
    overflow-y: auto; /* 添加垂直滾動條 */
    padding: 10px;
    border: 1px solid #ccc;
    border-radius: 5px;
    background-color: black;
    color: white;
}
"""



initial_system_prompt = """你是一個具備高級推理能力的AI助理。在面對問題或挑戰時，將你的思考過程分解為詳細且合乎邏輯的步驟。每一步應該：
1. 解釋你的推理
2. 考慮可能的缺陷或替代觀點
3. 在必要時調整你的思考

使用以下的Markdown結構：

## 初步想法
[對問題初步解決方法的簡短概述]

## 詳細推理
1. [第一步]
   **解釋:** [詳細的描述這個步驟]
   **可能問題:** [考慮有什麼可能的錯誤，或者有誤解的地方]
   **改進:** [基於可能的問題，你可以怎樣調整這個步驟]

2. [第二步]
   **解釋:** [詳細解釋這個步驟]
   **可能問題:** [考慮有沒什有什麼可能的錯誤或者誤解]
   **改進:** [基於可能的問題，你可以怎樣調整這個步驟]

...

## 初步的答案
[基於此推理，您目前的最佳答案]

## 信心和局限性
[討論您對這個答案的信心以及仍然存在哪些限製或不確定性]

全面、自我批評，並清楚表達你的推理."""

followup_system_prompt = """你被賦予了一項任務，作為一個AI助理，你需要批判性地分析並改進之前的問題解決步驟。
審視原始查詢和之前的推理過程，然後提供新的觀點或更深入的分析. 使用以下的Markdown結構：

## 批判性分析
[對之前推理進行詳細評估，強調其優點和不足之處]

## 替代方法
1. [新方法或精煉方法的第一步]
   **解釋:** [詳細解釋這一步，如果相關的話可以參考之前的推理]
   **改進措施:** [這一步如何改進或與之前的思考有所不同]

2. [新方法或精煉方法的第二步]
   **解釋:** [詳細解釋這一步是如何建立在或與之前的思考不同的]
   **改進措施:** [這一步如何改進或與之前的思考有所不同]

...

## 修改後的答案
[根據這一新分析得出的更新答案]

## 比較優勢
[解釋為什麼這個新方法可能比之前的更好，或者它帶來了哪些獨特的洞察]

在進行批判時，既要指出之前的不足之處，也要提出具體的改進建議。
通過提供新的洞察和思路，可以幫助提升問題解決的效果。
每次轉變都應旨在提高整個推理過程的品質，使其更加完善和有效."""

synthesis_prompt = """你被賦予了一項任務，作為一個AI助理，
你需要將多個推理步驟綜合成最終的全面答案。你會收到解決某一問題的幾個推理步驟。
你的任務是將這些步驟整合在一起，形成一個完整且全面的答案:

1. 分析每個回合，考慮其優點、弱點和獨特貢獻.
2. 比較和對比不同的方法，注意推理是如何演變的.
3. 將所有回合的見解綜合起來，形成一個最終、合乎邏輯的答案.
4. 提供詳細解釋和簡潔明瞭的最終答案.

使用以下的Markdown結構：

## 推理轉折的分析
[請對每個推理步驟進行簡要分析，突出關鍵洞察和發展]

## 思想演變
[描述推理過程在各個步驟中的演化，註明主要的轉折點和改進]

## 合成推理
[最終合成推理過程應結合各個步驟中的最佳洞察和發展]

## 綜合最終答案
[詳細且周密的答案，涵蓋了所有相關的見解]

## 簡潔的答案
[簡潔明瞭的最終答案，適合一般讀者理解，不超過2至3句話.]

## 信心評估
[評估最終答案的信心水準，討論任何剩餘的不確定性和限制因素]

在分析中要詳盡周全，推理過程要清晰明瞭，並且要注意最終結論的優勢和局限性."""



def get_ollama_models():
    try:
        # 假设 Ollama 提供一个 API 端点来获取模型列表
        response = requests.get("http://localhost:11434/api/tags")  # 替换为实际 API 地址
        response.raise_for_status()  # 确保请求成功
        models = response.json()  # 假设返回的是一个 JSON 列表
        return models.get("models", [])  # 假设模型列表在 "models" 字段中
    except Exception as e:
        return [f"Error: {str(e)}"]




# 定義與 Ollama 串流互動的函數
def call_ollama(prompt, system):
    # 設定請求的資料
    
    for i in range(1):
        
        headers = {"Content-Type": "application/json"}
        data = {
            "model": MODEL_NAME,
            "prompt": prompt,
            "system": system,
            "stream": True
        }
            
        # 發送 POST 請求給 Ollama API
        response = requests.post(OLLAMA_API_URL, headers=headers, json=data, stream=True)
    
        # 處理串流回應
        if response.status_code == 200:
            for line in response.iter_lines():
                if line.strip():  # 確保不是空行
                    yield line.decode("utf-8")  # line 已是解碼後的字串
        else:
            yield f"Error: {response.status_code}, {response.text}"

# 定義 Gradio 界面的回應函數
def chatbot_interface(prompt, system):
    for chunk in call_ollama (prompt, system):
        #print(f"The chunk is {chunk}")
        jsn = json.loads(chunk)        
        rep = jsn['response']  
        
        yield rep  # 每次更新內容


def synthesize_turns(query: str, turns: list) -> str:
    global this_turn
    """Synthesize multiple turns of reasoning into a final answer."""
    turns_text = "\n\n".join([f"Turn {i+1}:\n{turn}" for i, turn in enumerate(turns)])
    prompt = f"Original Query: {query}\n\nTurns of Reasoning:\n{turns_text}\n\n將這些步驟綜合為一個最終的、全面的答案。"
    for chunk in chatbot_interface(prompt, synthesis_prompt):
        this_turn += chunk
        yield chunk

def generate_turn(query: str, previous_turns: list = None) -> str:
    global this_turn
    """Generate a single turn of reasoning, considering previous turns if available."""
    is_first_turn = previous_turns is None or len(previous_turns) == 0
    if is_first_turn:
        system = initial_system_prompt
        prompt = f"Query: {query}\n\n提供您對此問題的初步想法和詳細推理。"
    else:
        system = followup_system_prompt
        previous_content = "\n\n".join(previous_turns)
        prompt = f"Original Query: {query}\n\nPrevious Turns:\n{previous_content}\n\n提供下一步的推理，批判性地分析之前的步驟並提供新的見解或改進建議。"

    for chunk in chatbot_interface(prompt, system):
        this_turn += chunk
        yield chunk

        
def full_cot_reasoning(query, turn_count_num, model):
    global MODEL_NAME
    """Perform enhanced Chain of Thought reasoning with multiple turns and synthesis."""
    if len(model) != 0:
        MODEL_NAME=model
    if len(turn_count_num) == 0:
        turn_count_num="3"
    full_output = f"使用{MODEL_NAME}模型進行推理\n\n"
    total_turns=int(turn_count_num)
    start_time = time.time()
    turns = []
    turn_times = []
    full_output += f"# 加強推理思考鏈\n\n## 原問題\n{query}\n\n"

    for i in range(total_turns):  
        this_turn=""
        full_output += f"\n## 第 {i+1} 回\n\n"
        turn_start = time.time()
        for chunk in generate_turn(query, turns):
            full_output += chunk
            yield full_output
        turns.append(this_turn)
        turn_times.append(time.time() - turn_start)
        

    mid_time = time.time()
    full_output += f"## 最終綜合\n\n"
    for chunk in synthesize_turns(query, turns):
        full_output += chunk
        yield full_output
    end_time = time.time()

    timing = {
        'turn_times': turn_times,
        'total_turns_time': mid_time - start_time,
        'synthesis_time': end_time - mid_time,
        'total_time': end_time - start_time
    }

    full_output += f"## 耗用時間\n"
    for i in range(total_turns):
        full_output += f"- 第{i+1}回: {timing['turn_times'][0]:.2f}s\n"
    full_output += f"- 總思考時間: {timing['total_turns_time']:.2f}s\n"
    full_output += f"- 結論耗時: {timing['synthesis_time']:.2f}s\n"
    full_output += f"- 總時間: {timing['total_time']:.2f}s\n"

    yield full_output

    
_models = get_ollama_models()
models=[]
for m in _models:
    models.append(m["name"])

# 建立 Gradio 介面
with gr.Blocks(css=custom_css) as demo:
    with gr.Row():
        with gr.Column():
            with gr.Row():
                prompt_input = gr.Textbox(label="輸入您的問題", placeholder="輸入提示...")
                turn_count   = gr.Textbox(label="思考的輪數", placeholder="3")
                model=gr.Dropdown(choices=models, label="使用模型", value=MODEL_NAME)
            submit_button = gr.Button("提交")
    with gr.Row():
        chat_output = gr.Markdown(elem_id="scrollable-markdown")



    submit_button.click(full_cot_reasoning, inputs=[prompt_input,turn_count, model], outputs=chat_output)


# 啟動 Gradio 介面
demo.launch()


